from abc import abstractmethod

import requests
import jsonpickle
import json

HTTP_OK = 200

class CommandBase(object):
    def __init__(self):
        self._url = ""

    @property
    def _method(self):
        return "PUT"

    @property
    def url(self):
        return self._url
    
    @url.setter
    def url(self, new_url):
        self._url = new_url

    def get_url(self):
        return self._url

    def get_method(self):
        return self._method

    @abstractmethod
    def process_response(self, json_obj):
        return None


class InfoCommandBase(CommandBase):
    def __init__(self):
        super(InfoCommandBase, self).__init__()

    @property
    def _method(self):
        return "GET"
    
    @property
    def url(self):
        return CommandBase.url.fget(self)

    @url.setter
    def url(self, new_url):
        CommandBase.url.fset(self, new_url)


class ProtoConstants(object):
    ACTION_MODIFY = "MODIFY"
    HEADER_AUTH = "AUTH"
    STATUS_SUCCESS = "success"
    RESPONSE_ITEM = "item"
    RESPONSE_ITEMS = "items"
    AUTH_TOKEN = "auth_token"
    CHALLENGE_TYPE = "challenge_type"
    PAIRING_REQ_TOKEN = "pairing_req_token"

    class Item(object):
        HASHVAL = "hashval"
        CNAME = "cname"
        TYPE = "type"
        NAME = "name"
        VALUE = "value"
        METADATA = "metadata"
        ELEMENTS = "elements"


class KeyCodes(object):
    CODES = {
        "tv": 
        {
            "SEEK_FWD":     (2, 0),
            "SEEK_BACK":    (2, 1),
            "PAUSE":        (2, 2),
            "PLAY":         (2, 3),
            "DOWN":         (3, 0),
            "LEFT":         (3, 1),
            "OK":           (3, 2),
            "UP":           (3, 3),
            "LEFT2":        (3, 4),
            "RIGHT":        (3, 5),
            "BACK":         (4, 0),
            "SMARTCAST":    (4, 3),
            "CC_TOGGLE":    (4, 4),
            "INFO":         (4, 6),
            "MENU":         (4, 8),
            "HOME":         (4, 15),
            "VOL_DOWN":     (5, 0),
            "VOL_UP":       (5, 1),
            "MUTE_OFF":     (5, 2),
            "MUTE_ON":      (5, 3),
            "MUTE_TOGGLE":  (5, 4),
            "PIC_MODE":     (6, 0),
            "PIC_SIZE":     (6, 2),
            "INPUT_NEXT":   (7, 1),
            "CH_DOWN":      (8, 0),
            "CH_UP":        (8, 1),
            "CH_PREV":      (8, 2),
            "EXIT":         (9, 0),
            "POW_OFF":      (11, 0),
            "POW_ON":       (11, 1),
            "POW_TOGGLE":   (11, 2)
        },
        "soundbar": 
        {
            "PAUSE":        (2, 2),
            "PLAY":         (2, 3),
            "VOL_DOWN":     (5, 0),
            "VOL_UP":       (5, 1),
            "MUTE_OFF":     (5, 2),
            "MUTE_ON":      (5, 3),
            "MUTE_TOGGLE":  (5, 4),
            "POW_OFF":      (11, 0),
            "POW_ON":       (11, 1),
            "POW_TOGGLE":   (11, 2)
        }
    }

    class KeyPressActions(object):
        KEY_DOWN = "KEYDOWN"
        KEY_UP = "KEYUP"
        KEY_PRESS = "KEYPRESS"


class Endpoints(object):
    ENDPOINTS = {
        "tv":
        {
            "BEGIN_PAIR":   "/pairing/start",
            "FINISH_PAIR":  "/pairing/pair",
            "CANCEL_PAIR":  "/pairing/cancel",
            "INPUTS":       "/menu_native/dynamic/tv_settings/devices/name_input",
            "CURR_INPUT":   "/menu_native/dynamic/tv_settings/devices/current_input",
            "SET_INPUT":   "/menu_native/dynamic/tv_settings/devices/current_input",
            "POWER":        "/state/device/power_mode",
            "KEY_PRESS":    "/key_command/",
            "VOLUME":       "/menu_native/dynamic/tv_settings/audio/volume"
        },
        "soundbar":
        {
            "BEGIN_PAIR":   "/pairing/start",
            "FINISH_PAIR":  "/pairing/pair",
            "CANCEL_PAIR":  "/pairing/cancel",
            "INPUTS":       "/menu_native/dynamic/audio_settings/input",
            "CURR_INPUT":   "/menu_native/dynamic/audio_settings/input/current_input",
            "SET_INPUT":   "/menu_native/dynamic/audio_settings/input/current_input",
            "POWER":        "/state/device/power_mode",
            "KEY_PRESS":    "/key_command/",
            "VOLUME":       "/menu_native/dynamic/audio_settings/audio/volume"
        }
    }


class CNames(object):
    class Audio(object):
        VOLUME = "volume"


def get_json_obj(json_obj, key):
    key = key.lower()
    for k, v in json_obj.items():
        if k.lower() == key:
            return v
    return None


def validate_response(web_response):
    if HTTP_OK != web_response.status_code:
        raise Exception("TV is unreachable? Status code: {0}".format(web_response.status_code))
    try:
        data = json.loads(web_response.text)
    except:
        raise Exception("Failed to parse response: {0}".format(web_response.content))
    status_obj = get_json_obj(data, "status")
    if status_obj is None:
        raise Exception("Unknown response")
    result_status = get_json_obj(status_obj, "result")
    if result_status is None or result_status.lower() != ProtoConstants.STATUS_SUCCESS:
        raise Exception("Unexpected response {0}: {1}".format(result_status, get_json_obj(status_obj, "detail")))
    return data


def invoke_api(ip, command, logger, headers=None):
    if headers is None:
        headers = {}

    try:
        method = command.get_method()
        if ":" not in ip:
            ip = ip + ":7345"
        url = "https://{0}{1}".format(ip, command.get_url())
        data = jsonpickle.encode(command, unpicklable=False)
        if "get" == method.lower():
            response = requests.get(url=url, headers=headers, verify=False)
        else:
            headers["Content-Type"] = "application/json"
            response = requests.request(method=method, data=str(data), url=url, headers=headers, verify=False)
        json_obj = validate_response(response)
        return command.process_response(json_obj)
    except Exception as e:
        logger.error("Failed to execute command: %s", e)
        return None


def invoke_api_auth(ip, command, auth_token, logger):
    headers = {ProtoConstants.HEADER_AUTH: auth_token}
    return invoke_api(ip, command, logger, headers)



import logging
from urllib.parse import urlsplit

import requests
import xmltodict

from .cmd_input import GetInputsListCommand, GetCurrentInputCommand, ChangeInputCommand
from .cmd_pair import BeginPairCommand, CancelPairCommand, PairChallengeCommand
from .cmd_power import GetPowerStateCommand
from .cmd_remote import EmulateRemoteCommand
from .cmd_settings import GetCurrentAudioCommand
from .discovery import discover
from .protocol import invoke_api, invoke_api_auth, KeyCodes

_LOGGER = logging.getLogger(__name__)


class DeviceDescription(object):
    def __init__(self, ip, name, model, udn):
        self.ip = ip
        self.name = name
        self.model = model
        self.udn = udn


class Vizio(object):
    def __init__(self, device_id, ip, name, auth_token="", device_type="tv"):
        self._device_type = device_type.lower()
        if self._device_type != "tv" and self._device_type != "soundbar":
            raise Exception("Invalid device type specified. Use either 'tv' or 'soundbar'")
        
        self._ip = ip
        self._name = name
        self._device_id = device_id
        self._auth_token = auth_token

    def __invoke_api(self, cmd):
        return invoke_api(self._ip, cmd, _LOGGER)

    def __invoke_api_may_need_auth(self, cmd):
        if self._auth_token is None or "" == self._auth_token:
            if self._device_type == "soundbar":
                return invoke_api(self._ip, cmd, _LOGGER)
            else:
                raise Exception("Empty auth token. To target a soundbar and bypass auth requirements, pass 'soundbar' as device_type")
        return invoke_api_auth(self._ip, cmd, self._auth_token, _LOGGER)


    def __remote(self, key_list):
        key_codes = []
        if isinstance(key_list, list) is False:
            key_list = [key_list]

        for key in key_list:
            if key not in KeyCodes.CODES[self._device_type]:
                _LOGGER.error("Key Code of '%s' not found for device type of '%s'", key, self._device_type)
                return False
            else:
                key_codes.append(KeyCodes.CODES[self._device_type][key])
        cmd = EmulateRemoteCommand(key_codes, self._device_type)
        result = self.__invoke_api_may_need_auth(cmd)
        return result is not None

    def __remote_multiple(self, key_code, num):
        key_codes = []
        for ii in range(0, num):
            key_codes.append(key_code)
        return self.__remote(key_codes)

    @staticmethod
    def discovery():
        results = []
        devices = discover("urn:dial-multiscreen-org:device:dial:1")
        for dev in devices:
            data = xmltodict.parse(requests.get(dev.location, verify=False).text)

            if "root" not in data or "device" not in data["root"]:
                continue

            root = data["root"]["device"]
            manufacturer = root["manufacturer"]
            if manufacturer is None or "VIZIO" != manufacturer:
                continue
            split_url = urlsplit(dev.location)
            device = DeviceDescription(split_url.hostname, root["friendlyName"], root["modelName"], root["UDN"])
            results.append(device)

        return results

    def start_pair(self):
        return self.__invoke_api(BeginPairCommand(self._device_id, self._name, self._device_type))

    def stop_pair(self):
        return self.__invoke_api(CancelPairCommand(self._device_id, self._name, self._device_type))

    def pair(self, ch_type, token, pin):
        return self.__invoke_api(PairChallengeCommand(self._device_id, ch_type, token, pin, self._device_type))

    def get_inputs(self):
        return self.__invoke_api_may_need_auth(GetInputsListCommand(self._device_type))

    def get_current_input(self):
        return self.__invoke_api_may_need_auth(GetCurrentInputCommand(self._device_type))

    def get_power_state(self):
        return self.__invoke_api_may_need_auth(GetPowerStateCommand(self._device_type))

    def pow_on(self):
        return self.__remote("POW_ON")

    def pow_off(self):
        return self.__remote("POW_OFF")

    def pow_toggle(self):
        return self.__remote("POW_TOGGLE")

    def vol_up(self, num=1):
        return self.__remote_multiple("VOL_UP", num)

    def vol_down(self, num=1):
        return self.__remote_multiple("VOL_DOWN", num)

    def get_current_volume(self):
        return self.__invoke_api_may_need_auth(GetCurrentAudioCommand(self._device_type))

    def ch_up(self, num=1):
        return self.__remote_multiple("CH_UP", num)

    def ch_down(self, num=1):
        return self.__remote_multiple("CH_DOWN", num)

    def ch_prev(self):
        return self.__remote("CH_PREV")

    def mute_on(self):
        return self.__remote("MUTE_ON")

    def mute_off(self):
        return self.__remote("MUTE_OFF")

    def mute_toggle(self):
        return self.__remote("MUTE_TOGGLE")

    def input_next(self):
        # HACK: Single call just invoking overlay menu with current input
        return self.__remote_multiple("INPUT_NEXT", 2)

    def input_switch(self, name):
        cur_input = self.get_current_input()
        if cur_input is None:
            _LOGGER.error("Couldn't detect current input")
            return False
        return self.__invoke_api_may_need_auth(ChangeInputCommand(cur_input.id, name, self._device_type))

    def play(self):
        return self.__remote("PLAY")

    def pause(self):
        return self.__remote("PAUSE")

    def remote(self, key):
        return self.__remote(key)

    def get_device_keys(self):
        return KeyCodes.CODES[self._device_type].keys()



import logging
import click
import sys
import pyvizio

if sys.version_info < (3, 4):
    print("To use this script you need python 3.4 or newer, got %s" %
          sys.version_info)
    sys.exit(1)

_LOGGER = logging.getLogger(__name__)
DEVICE_ID = "pyvizio"
DEVICE_NAME = "Python Vizio"

pass_vizio = click.make_pass_decorator(pyvizio.Vizio)


@click.group(invoke_without_command=False)
@click.option('--ip', envvar="VIZIO_IP", required=True, help="IP of the device to connect to (optionally add custom port by specifying '<IP>:<PORT>')")
@click.option('--auth', envvar="VIZIO_AUTH", required=False, help="Auth token for the device to connect to (refer to documentation on how to obtain auth token)")
@click.option("--device_type", envvar="VIZIO_DEVICE_TYPE", required=False, default="tv", type=click.Choice(["tv", "soundbar"]))
@click.pass_context
def cli(ctx, ip, auth, device_type):
    logging.basicConfig(level=logging.INFO)
    ctx.obj = pyvizio.Vizio(DEVICE_ID, ip, DEVICE_NAME, auth, device_type)


@cli.command()
def discover():
    logging.basicConfig(level=logging.INFO)
    devices = pyvizio.Vizio.discovery()
    log_data = "Available devices:" \
               "\nIP\tModel\tFriendly name"
    for dev in devices:
        log_data += "\n{0}\t{1}\t{2}".format(dev.ip, dev.model, dev.name)
    _LOGGER.info(log_data)


@cli.command()
@pass_vizio
def pair(vizio):
    _LOGGER.info("Initiating pairing process, please check your device for pin upon success")
    pair_data = vizio.start_pair()
    if pair_data is not None:
        _LOGGER.info("Challenge type: %s", pair_data.ch_type)
        _LOGGER.info("Challenge token: %s", pair_data.token)


@cli.command()
@pass_vizio
def pair_stop(vizio):
    _LOGGER.info("Sending stop pair command")
    vizio.stop_pair()


@cli.command()
@click.option('--ch_type', required=False, default=1, help="Challenge type obtained from pair command")
@click.option('--token', required=True, help="Challenge token obtained from pair command")
@click.option('--pin', required=True, help="PIN obtained from device after running pair command")
@pass_vizio
def pair_finish(vizio, ch_type, token, pin):
    _LOGGER.info("Finishing pairing")
    pair_data = vizio.pair(ch_type, token, pin)
    if pair_data is not None:
        _LOGGER.info("Authorization token: %s", pair_data.auth_token)


@cli.command()
@pass_vizio
def input_list(vizio):
    inputs = vizio.get_inputs()
    if inputs:
        log_data = "Available inputs:" \
                "\nName\tFriendly name\tType\tID"
        for v_input in inputs:
            log_data += "\n{0}\t{1}\t{2}\t{3}".format(v_input.name, v_input.meta_name, v_input.type, v_input.id)

        _LOGGER.info(log_data)
    else:
        _LOGGER.error("Couldn't get available inputs")


@cli.command()
@pass_vizio
def input_get(vizio):
    data = vizio.get_current_input()
    if data:
        _LOGGER.info("Current input: %s", data.meta_name)
    else:
        _LOGGER.error("Couldn't get current input")
    

@cli.command()
@pass_vizio
def power_get(vizio):
    is_on = vizio.get_power_state()
    _LOGGER.info("Device is on" if is_on else "Device is off")


@cli.command()
@click.argument("state", required=False, default="toggle", type=click.Choice(["toggle", "on", "off"]))
@pass_vizio
def power(vizio, state):
    if "on" == state:
        txt = "Turning ON"
        result = vizio.pow_on()
    elif "off" == state:
        txt = "Turning OFF"
        result = vizio.pow_off()
    else:
        txt = "Toggling power"
        result = vizio.pow_toggle()
    _LOGGER.info(txt)
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument("state", required=False, default="up", type=click.Choice(["up", "down"]))
@click.argument("amount", required=False, default=1, type=click.IntRange(1, 100, clamp=True))
@pass_vizio
def volume(vizio, state, amount):
    amount = int(amount)
    if "up" == state:
        txt = "Increasing volume"
        result = vizio.vol_up(amount)
    else:
        txt = "Decreasing volume"
        result = vizio.vol_down(amount)
    _LOGGER.info(txt)
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@pass_vizio
def volume_current(vizio):
    _LOGGER.info("Current volume: %s", vizio.get_current_volume())


@cli.command()
@click.argument("state", required=False, default="previous", type=click.Choice(["up", "down", "previous"]))
@click.argument("amount", required=False, default=1, type=click.IntRange(1, 100, clamp=True))
@pass_vizio
def channel(vizio, state, amount):
    amount = int(amount)
    if "up" == state:
        txt = "Channel up"
        result = vizio.ch_up(amount)
    elif "down" == state:
        txt = "Channel down"
        result = vizio.ch_down(amount)
    else:
        txt = "Previous channel"
        result = vizio.ch_prev()
    _LOGGER.info(txt)
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument("state", required=False, default="toggle", type=click.Choice(["toggle", "on", "off"]))
@pass_vizio
def mute(vizio, state):
    if "on" == state:
        txt = "Muting"
        result = vizio.mute_on()
    elif "off" == state:
        txt = "Un-muting"
        result = vizio.mute_off()
    else:
        txt = "Toggling mute"
        result = vizio.mute_toggle()
    _LOGGER.info(txt)
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@pass_vizio
def input_next(vizio):
    result = vizio.input_next()
    _LOGGER.info("Circling input")
    _LOGGER.info("OK" if result else "ERROR")


@cli.command(name="input")
@click.option('--name', required=True)
@pass_vizio
def input(vizio, name):
    result = vizio.input_switch(name)
    _LOGGER.info("Switching input")
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@pass_vizio
def play(vizio):
    result = vizio.play()
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@pass_vizio
def pause(vizio):
    result = vizio.pause()
    _LOGGER.info("OK" if result else "ERROR")


@cli.command()
@click.argument("key", required=True)
@pass_vizio
def key_press(vizio, key):
    _LOGGER.info("Emulating pressing of '%s' key", key)
    result = vizio.remote(key)
    _LOGGER.info("OK" if result else "ERROR")


if __name__ == "__main__":
    cli()



from setuptools import setup

with open('pyvizio/version.py') as f: exec(f.read())
with open('README.md', 'r') as myfile:
    longdescription=myfile.read()

setup(
    name='pyvizio',

    version=__version__,
    description='Python library for interfacing with Vizio SmartCast TVs and Sound Bars (2016+ models)',
    long_description=longdescription,
    long_description_content_type="text/markdown",
    url='https://github.com/vkorn/pyvizio',

    author='Vlad Korniev',
    author_email='vladimir.kornev@gmail.com',

    license='GPLv3',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
    ],

    keywords='vizio smartcast',

    packages=["pyvizio"],

    install_requires=['click', 'requests', 'jsonpickle', 'xmltodict'],
    entry_points={
        'console_scripts': [
            'pyvizio=pyvizio.cli:cli',
        ],
    },
)
